@@ -11,7 +11,8 @@ namespace mold {
 
 struct X86_64;
 struct I386;
-struct ARM64;
+struct ARM64LE;
+struct ARM64BE;
 struct ARM32;
 struct RV64LE;
 struct RV64BE;
@@ -23,7 +24,8 @@ struct PPC64V2;
 struct S390X;
 struct SPARC64;
 struct M68K;
-struct SH4;
+struct SH4LE;
+struct SH4BE;
 struct LOONGARCH64;
 struct LOONGARCH32;
 
@@ -49,10 +51,6 @@ std::ostream &operator<<(std::ostream &out, const ElfRel<E> &rel) {
   return out;
 }
 
-enum : u32 {
-  R_NONE = 0,
-};
-
 enum : u32 {
   SHN_UNDEF = 0,
   SHN_LORESERVE = 0xff00,
@@ -296,6 +294,7 @@ enum : u32 {
   DT_VERNEEDNUM = 0x6fffffff,
   DT_PPC_GOT = 0x70000000,
   DT_PPC64_GLINK = 0x70000000,
+  DT_RISCV_VARIANT_CC = 0x70000001,
   DT_AARCH64_VARIANT_PCS = 0x70000005,
   DT_AUXILIARY = 0x7ffffffd,
   DT_FILTER = 0x7fffffff,
@@ -408,6 +407,10 @@ enum : u32 {
 // Relocation types
 //
 
+enum : u32 {
+  R_NONE = 0,
+};
+
 enum : u32 {
   R_X86_64_NONE = 0,
   R_X86_64_64 = 1,
@@ -783,10 +786,6 @@ enum : u32 {
   R_RISCV_ALIGN = 43,
   R_RISCV_RVC_BRANCH = 44,
   R_RISCV_RVC_JUMP = 45,
-  R_RISCV_RVC_LUI = 46,
-  R_RISCV_GPREL_LO12_I = 47,
-  R_RISCV_GPREL_LO12_S = 48,
-  R_RISCV_GPREL_HI20 = 49,
   R_RISCV_RELAX = 51,
   R_RISCV_SUB6 = 52,
   R_RISCV_SET6 = 53,
@@ -1350,6 +1349,16 @@ enum : u32 {
   R_LARCH_TLS_DESC_PCREL20_S2 = 126,
 };
 
+// Returns true if a given relocation is of type used for direct
+// function call.
+template <typename E>
+inline bool is_func_call_rel(const ElfRel<E> &r) {
+  for (u32 r_type : E::R_FUNCALL)
+    if (r.r_type == r_type)
+      return true;
+  return false;
+}
+
 //
 // DWARF data types
 //
@@ -1464,7 +1473,7 @@ template <typename E> using U32 = std::conditional_t<E::is_le, ul32, ub32>;
 template <typename E> using U64 = std::conditional_t<E::is_le, ul64, ub64>;
 
 template <typename E> using Word = std::conditional_t<E::is_64, U64<E>, U32<E>>;
-template <typename E> using IWord = std::conditional_t<E::is_64, I64<E>, I32<E>>;
+template <typename E> using SWord = std::conditional_t<E::is_64, I64<E>, I32<E>>;
 
 template <typename E> requires E::is_64
 struct ElfSym<E> {
@@ -1474,18 +1483,41 @@ struct ElfSym<E> {
   bool is_weak() const { return st_bind == STB_WEAK; }
   bool is_undef_weak() const { return is_undef() && is_weak(); }
 
+  bool ppc64_preserves_r2() const { return ppc64_local_entry != 1; }
+  bool ppc64_uses_toc() const { return ppc64_local_entry > 1; }
+
   U32<E> st_name;
 
 #ifdef __LITTLE_ENDIAN__
   u8 st_type : 4;
   u8 st_bind : 4;
-  u8 st_visibility : 2;
-  u8 : 6;
+  union {
+    u8 st_visibility : 2;
+    struct {
+      u8 : 7;
+      u8 arm64_variant_pcs : 1;
+    };
+    struct {
+      u8 : 7;
+      u8 riscv_variant_cc : 1;
+    };
+    struct {
+      u8 : 5;
+      u8 ppc64_local_entry : 3;
+    };
+  };
 #else
   u8 st_bind : 4;
   u8 st_type : 4;
-  u8 : 6;
-  u8 st_visibility : 2;
+  union {
+    struct {
+      u8 : 6;
+      u8 st_visibility : 2;
+    };
+    u8 arm64_variant_pcs : 1;
+    u8 riscv_variant_cc : 1;
+    u8 ppc64_local_entry : 3;
+  };
 #endif
 
   U16<E> st_shndx;
@@ -1508,13 +1540,23 @@ struct ElfSym<E> {
 #ifdef __LITTLE_ENDIAN__
   u8 st_type : 4;
   u8 st_bind : 4;
-  u8 st_visibility : 2;
-  u8 : 6;
+  union {
+    u8 st_visibility : 2;
+    struct {
+      u8 : 7;
+      u8 riscv_variant_cc : 1;
+    };
+  };
 #else
   u8 st_bind : 4;
   u8 st_type : 4;
-  u8 : 6;
-  u8 st_visibility : 2;
+  union {
+    struct {
+      u8 : 6;
+      u8 st_visibility : 2;
+    };
+    u8 riscv_variant_cc : 1;
+  };
 #endif
 
   U16<E> st_shndx;
@@ -1588,7 +1630,7 @@ struct ElfPhdr<E> {
 // We don't want to have too many `if (REL)`s and `if (RELA)`s in our
 // codebase, so ElfRel always takes r_addend as a constructor argument.
 // If it's REL, the argument will simply be ignored.
-template <typename E> requires E::is_le && E::is_rela
+template <typename E> requires (E::is_le && E::is_rela)
 struct ElfRel<E> {
   ElfRel() = default;
   ElfRel(u64 offset, u32 type, u32 sym, i64 addend)
@@ -1597,10 +1639,10 @@ struct ElfRel<E> {
   Word<E> r_offset;
   std::conditional_t<E::is_64, U32<E>, u8> r_type;
   std::conditional_t<E::is_64, U32<E>, U24<E>> r_sym;
-  IWord<E> r_addend;
+  SWord<E> r_addend;
 };
 
-template <typename E> requires (!E::is_le) && E::is_rela
+template <typename E> requires (!E::is_le && E::is_rela)
 struct ElfRel<E> {
   ElfRel() = default;
   ElfRel(u64 offset, u32 type, u32 sym, i64 addend)
@@ -1609,10 +1651,10 @@ struct ElfRel<E> {
   Word<E> r_offset;
   std::conditional_t<E::is_64, U32<E>, U24<E>> r_sym;
   std::conditional_t<E::is_64, U32<E>, u8> r_type;
-  IWord<E> r_addend;
+  SWord<E> r_addend;
 };
 
-template <typename E> requires E::is_le && (!E::is_rela)
+template <typename E> requires (E::is_le && !E::is_rela)
 struct ElfRel<E> {
   ElfRel() = default;
   ElfRel(u64 offset, u32 type, u32 sym, i64 addend = 0)
@@ -1623,7 +1665,7 @@ struct ElfRel<E> {
   std::conditional_t<E::is_64, U32<E>, U24<E>> r_sym;
 };
 
-template <typename E> requires (!E::is_le) && (!E::is_rela)
+template <typename E> requires (!E::is_le && !E::is_rela)
 struct ElfRel<E> {
   ElfRel() = default;
   ElfRel(u64 offset, u32 type, u32 sym, i64 addend = 0)
@@ -1634,16 +1676,6 @@ struct ElfRel<E> {
   std::conditional_t<E::is_64, U32<E>, u8> r_type;
 };
 
-// Returns true if a given relocation is of type used for direct
-// function call.
-template <typename E>
-inline bool is_func_call_rel(const ElfRel<E> &r) {
-  for (u32 r_type : E::R_FUNCALL)
-    if (r.r_type == r_type)
-      return true;
-  return false;
-}
-
 template <typename E>
 struct ElfDyn {
   Word<E> d_tag;
@@ -1711,67 +1743,6 @@ struct ElfNhdr {
 // Target-specific ELF data types
 //
 
-template <>
-struct ElfSym<ARM64> {
-  bool is_undef() const { return st_shndx == SHN_UNDEF; }
-  bool is_abs() const { return st_shndx == SHN_ABS; }
-  bool is_common() const { return st_shndx == SHN_COMMON; }
-  bool is_weak() const { return st_bind == STB_WEAK; }
-  bool is_undef_weak() const { return is_undef() && is_weak(); }
-
-  ul32 st_name;
-
-#ifdef __LITTLE_ENDIAN__
-  u8 st_type : 4;
-  u8 st_bind : 4;
-  u8 st_visibility : 2;
-  u8 : 5;
-  u8 arm64_variant_pcs : 1; // ARM64-specific
-#else
-  u8 st_bind : 4;
-  u8 st_type : 4;
-  u8 arm64_variant_pcs : 1;
-  u8 : 5;
-  u8 st_visibility : 2;
-#endif
-
-  ul16 st_shndx;
-  ul64 st_value;
-  ul64 st_size;
-};
-
-template <>
-struct ElfSym<PPC64V2> {
-  bool is_undef() const { return st_shndx == SHN_UNDEF; }
-  bool is_abs() const { return st_shndx == SHN_ABS; }
-  bool is_common() const { return st_shndx == SHN_COMMON; }
-  bool is_weak() const { return st_bind == STB_WEAK; }
-  bool is_undef_weak() const { return is_undef() && is_weak(); }
-
-  bool preserves_r2() const { return ppc_local_entry != 1; }
-  bool uses_toc() const { return ppc_local_entry > 1; }
-
-  ul32 st_name;
-
-#ifdef __LITTLE_ENDIAN__
-  u8 st_type : 4;
-  u8 st_bind : 4;
-  u8 st_visibility : 2;
-  u8 : 3;
-  u8 ppc_local_entry : 3; // This is PPC64V2-specific
-#else
-  u8 st_bind : 4;
-  u8 st_type : 4;
-  u8 ppc_local_entry : 3;
-  u8 : 3;
-  u8 st_visibility : 2;
-#endif
-
-  ul16 st_shndx;
-  ul64 st_value;
-  ul64 st_size;
-};
-
 template <>
 struct ElfRel<SPARC64> {
   ElfRel() = default;
@@ -1787,7 +1758,7 @@ struct ElfRel<SPARC64> {
 };
 
 template <>
-struct ElfRel<SH4> {
+struct ElfRel<SH4LE> {
   ElfRel() = default;
 
   // Addend is ignored except for base relocations because even though
@@ -1802,6 +1773,18 @@ struct ElfRel<SH4> {
   il32 r_addend;
 };
 
+template <>
+struct ElfRel<SH4BE> {
+  ElfRel() = default;
+  ElfRel(u64 offset, u32 type, u32 sym, i64 addend)
+    : r_offset(offset), r_sym(sym), r_type(type), r_addend(sym ? 0 : addend) {}
+
+  ub32 r_offset;
+  ub24 r_sym;
+  u8 r_type;
+  ib32 r_addend;
+};
+
 //
 // Machine descriptions
 //
@@ -1812,7 +1795,8 @@ template <typename E> concept needs_thunk = requires { E::thunk_size; };
 
 template <typename E> concept is_x86_64 = std::same_as<E, X86_64>;
 template <typename E> concept is_i386 = std::same_as<E, I386>;
-template <typename E> concept is_arm64 = std::same_as<E, ARM64>;
+template <typename E> concept is_arm64le = std::same_as<E, ARM64LE>;
+template <typename E> concept is_arm64be = std::same_as<E, ARM64BE>;
 template <typename E> concept is_arm32 = std::same_as<E, ARM32>;
 template <typename E> concept is_rv64le = std::same_as<E, RV64LE>;
 template <typename E> concept is_rv64be = std::same_as<E, RV64BE>;
@@ -1824,11 +1808,13 @@ template <typename E> concept is_ppc64v2 = std::same_as<E, PPC64V2>;
 template <typename E> concept is_s390x = std::same_as<E, S390X>;
 template <typename E> concept is_sparc64 = std::same_as<E, SPARC64>;
 template <typename E> concept is_m68k = std::same_as<E, M68K>;
-template <typename E> concept is_sh4 = std::same_as<E, SH4>;
+template <typename E> concept is_sh4le = std::same_as<E, SH4LE>;
+template <typename E> concept is_sh4be = std::same_as<E, SH4BE>;
 template <typename E> concept is_loongarch64 = std::same_as<E, LOONGARCH64>;
 template <typename E> concept is_loongarch32 = std::same_as<E, LOONGARCH32>;
 
 template <typename E> concept is_x86 = is_x86_64<E> || is_i386<E>;
+template <typename E> concept is_arm64 = is_arm64le<E> || is_arm64be<E>;
 template <typename E> concept is_arm = is_arm64<E> || is_arm32<E>;
 template <typename E> concept is_rv64 = is_rv64le<E> || is_rv64be<E>;
 template <typename E> concept is_rv32 = is_rv32le<E> || is_rv32be<E>;
@@ -1836,10 +1822,11 @@ template <typename E> concept is_riscv = is_rv64<E> || is_rv32<E>;
 template <typename E> concept is_ppc64 = is_ppc64v1<E> || is_ppc64v2<E>;
 template <typename E> concept is_ppc = is_ppc64<E> || is_ppc32<E>;
 template <typename E> concept is_sparc = is_sparc64<E>;
+template <typename E> concept is_sh4 = is_sh4le<E> || is_sh4be<E>;
 template <typename E> concept is_loongarch = is_loongarch64<E> || is_loongarch32<E>;
 
 struct X86_64 {
-  static constexpr std::string_view target_name = "x86_64";
+  static constexpr std::string_view name = "x86_64";
   static constexpr bool is_64 = true;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
@@ -1864,7 +1851,7 @@ struct X86_64 {
 };
 
 struct I386 {
-  static constexpr std::string_view target_name = "i386";
+  static constexpr std::string_view name = "i386";
   static constexpr bool is_64 = false;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = false;
@@ -1888,8 +1875,8 @@ struct I386 {
   static constexpr u32 R_FUNCALL[] = { R_386_PLT32 };
 };
 
-struct ARM64 {
-  static constexpr std::string_view target_name = "arm64";
+struct ARM64LE {
+  static constexpr std::string_view name = "arm64";
   static constexpr bool is_64 = true;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
@@ -1915,8 +1902,13 @@ struct ARM64 {
   static constexpr u32 R_FUNCALL[] = { R_AARCH64_JUMP26, R_AARCH64_CALL26 };
 };
 
+struct ARM64BE : ARM64LE {
+  static constexpr std::string_view name = "arm64be";
+  static constexpr bool is_le = false;
+};
+
 struct ARM32 {
-  static constexpr std::string_view target_name = "arm32";
+  static constexpr std::string_view name = "arm32";
   static constexpr bool is_64 = false;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = false;
@@ -1945,8 +1937,10 @@ struct ARM32 {
   };
 };
 
-struct RV64 {
+struct RV64LE {
+  static constexpr std::string_view name = "riscv64";
   static constexpr bool is_64 = true;
+  static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
   static constexpr u32 page_size = 4096;
   static constexpr u32 e_machine = EM_RISCV;
@@ -1968,18 +1962,15 @@ struct RV64 {
   static constexpr u32 R_FUNCALL[] = { R_RISCV_CALL, R_RISCV_CALL_PLT };
 };
 
-struct RV64LE : RV64 {
-  static constexpr std::string_view target_name = "riscv64";
-  static constexpr bool is_le = true;
-};
-
-struct RV64BE : RV64 {
-  static constexpr std::string_view target_name = "riscv64be";
+struct RV64BE : RV64LE {
+  static constexpr std::string_view name = "riscv64be";
   static constexpr bool is_le = false;
 };
 
-struct RV32 {
+struct RV32LE {
+  static constexpr std::string_view name = "riscv32";
   static constexpr bool is_64 = false;
+  static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
   static constexpr u32 page_size = 4096;
   static constexpr u32 e_machine = EM_RISCV;
@@ -2001,18 +1992,13 @@ struct RV32 {
   static constexpr u32 R_FUNCALL[] = { R_RISCV_CALL, R_RISCV_CALL_PLT };
 };
 
-struct RV32LE : RV32 {
-  static constexpr std::string_view target_name = "riscv32";
-  static constexpr bool is_le = true;
-};
-
-struct RV32BE : RV32 {
-  static constexpr std::string_view target_name = "riscv32be";
+struct RV32BE : RV32LE {
+  static constexpr std::string_view name = "riscv32be";
   static constexpr bool is_le = false;
 };
 
 struct PPC32 {
-  static constexpr std::string_view target_name = "ppc32";
+  static constexpr std::string_view name = "ppc32";
   static constexpr bool is_64 = false;
   static constexpr bool is_le = false;
   static constexpr bool is_rela = true;
@@ -2059,7 +2045,7 @@ struct PPC64 {
 };
 
 struct PPC64V1 : PPC64 {
-  static constexpr std::string_view target_name = "ppc64v1";
+  static constexpr std::string_view name = "ppc64v1";
   static constexpr bool is_le = false;
   static constexpr u32 plt_hdr_size = 44;
   static constexpr u32 pltgot_size = 0;
@@ -2069,7 +2055,7 @@ struct PPC64V1 : PPC64 {
 };
 
 struct PPC64V2 : PPC64 {
-  static constexpr std::string_view target_name = "ppc64v2";
+  static constexpr std::string_view name = "ppc64v2";
   static constexpr bool is_le = true;
   static constexpr u32 plt_hdr_size = 52;
   static constexpr u32 plt_size = 4;
@@ -2080,7 +2066,7 @@ struct PPC64V2 : PPC64 {
 };
 
 struct S390X {
-  static constexpr std::string_view target_name = "s390x";
+  static constexpr std::string_view name = "s390x";
   static constexpr bool is_64 = true;
   static constexpr bool is_le = false;
   static constexpr bool is_rela = true;
@@ -2104,7 +2090,7 @@ struct S390X {
 };
 
 struct SPARC64 {
-  static constexpr std::string_view target_name = "sparc64";
+  static constexpr std::string_view name = "sparc64";
   static constexpr bool is_64 = true;
   static constexpr bool is_le = false;
   static constexpr bool is_rela = true;
@@ -2128,7 +2114,7 @@ struct SPARC64 {
 };
 
 struct M68K {
-  static constexpr std::string_view target_name = "m68k";
+  static constexpr std::string_view name = "m68k";
   static constexpr bool is_64 = false;
   static constexpr bool is_le = false;
   static constexpr bool is_rela = true;
@@ -2150,8 +2136,8 @@ struct M68K {
   static constexpr u32 R_FUNCALL[] = { R_68K_PLT32 };
 };
 
-struct SH4 {
-  static constexpr std::string_view target_name = "sh4";
+struct SH4LE {
+  static constexpr std::string_view name = "sh4";
   static constexpr bool is_64 = false;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
@@ -2173,12 +2159,18 @@ struct SH4 {
   static constexpr u32 R_FUNCALL[] = { R_SH_PLT32 };
 };
 
+struct SH4BE : SH4LE {
+  static constexpr std::string_view name = "sh4be";
+  static constexpr bool is_le = false;
+  static constexpr u8 filler[] = { 0x90, 0x00 }; // nop
+};
+
 struct LOONGARCH64 {
-  static constexpr std::string_view target_name = "loongarch64";
+  static constexpr std::string_view name = "loongarch64";
   static constexpr bool is_64 = true;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
-  static constexpr u32 page_size = 16384;
+  static constexpr u32 page_size = 65536;
   static constexpr u32 e_machine = EM_LOONGARCH;
   static constexpr u32 plt_hdr_size = 32;
   static constexpr u32 plt_size = 16;
@@ -2199,11 +2191,11 @@ struct LOONGARCH64 {
 };
 
 struct LOONGARCH32 {
-  static constexpr std::string_view target_name = "loongarch32";
+  static constexpr std::string_view name = "loongarch32";
   static constexpr bool is_64 = false;
   static constexpr bool is_le = true;
   static constexpr bool is_rela = true;
-  static constexpr u32 page_size = 16384;
+  static constexpr u32 page_size = 65536;
   static constexpr u32 e_machine = EM_LOONGARCH;
   static constexpr u32 plt_hdr_size = 32;
   static constexpr u32 plt_size = 16;
