@@ -1,5 +1,4 @@
 #include "mold.h"
-#include "config.h"
 
 #include <cctype>
 #include <set>
@@ -408,51 +407,8 @@ void RelDynSection<E>::update_shdr(Context<E> &ctx) {
   this->shdr.sh_link = ctx.dynsym->shndx;
 }
 
-template <typename E>
-void RelDynSection<E>::sort(Context<E> &ctx) {
-  Timer t(ctx, "sort_dynamic_relocs");
-
-  ElfRel<E> *begin = (ElfRel<E> *)(ctx.buf + this->shdr.sh_offset);
-  ElfRel<E> *end = begin + this->shdr.sh_size / sizeof(ElfRel<E>);
-
-  auto get_rank = [](u32 r_type) {
-    if (r_type == E::R_RELATIVE)
-      return 0;
-    if constexpr (supports_ifunc<E>)
-      if (r_type == E::R_IRELATIVE)
-        return 2;
-    return 1;
-  };
-
-  // This is the reason why we sort dynamic relocations. Quote from
-  // https://www.airs.com/blog/archives/186:
-  //
-  //   The dynamic linker in glibc uses a one element cache when processing
-  //   relocs: if a relocation refers to the same symbol as the previous
-  //   relocation, then the dynamic linker reuses the value rather than
-  //   looking up the symbol again. Thus the dynamic linker gets the best
-  //   results if the dynamic relocations are sorted so that all dynamic
-  //   relocations for a given dynamic symbol are adjacent.
-  //
-  //   Other than that, the linker sorts together all relative relocations,
-  //   which don't have symbols. Two relative relocations, or two relocations
-  //   against the same symbol, are sorted by the address in the output
-  //   file. This tends to optimize paging and caching when there are two
-  //   references from the same page.
-  //
-  // We group IFUNC relocations at the end of .rel.dyn because we want to
-  // apply all the other relocations before running user-supplied ifunc
-  // resolver functions.
-  tbb::parallel_sort(begin, end, [&](const ElfRel<E> &a, const ElfRel<E> &b) {
-    return std::tuple(get_rank(a.r_type), a.r_sym, a.r_offset) <
-           std::tuple(get_rank(b.r_type), b.r_sym, b.r_offset);
-  });
-}
-
 template <typename E>
 void RelrDynSection<E>::update_shdr(Context<E> &ctx) {
-  this->shdr.sh_link = ctx.dynsym->shndx;
-
   i64 n = 0;
   for (Chunk<E> *chunk : ctx.chunks)
     n += chunk->relr.size();
@@ -537,11 +493,10 @@ void ShstrtabSection<E>::copy_buf(Context<E> &ctx) {
 
 template <typename E>
 i64 DynstrSection<E>::add_string(std::string_view str) {
-  if (this->shdr.sh_size == 0)
+  if (this->shdr.sh_size == 0) {
+    strings.insert({"", 0});
     this->shdr.sh_size = 1;
-
-  if (str.empty())
-    return 0;
+  }
 
   auto [it, inserted] = strings.insert({str, this->shdr.sh_size});
   if (inserted)
@@ -551,9 +506,6 @@ i64 DynstrSection<E>::add_string(std::string_view str) {
 
 template <typename E>
 i64 DynstrSection<E>::find_string(std::string_view str) {
-  if (str.empty())
-    return 0;
-
   auto it = strings.find(str);
   assert(it != strings.end());
   return it->second;
@@ -562,12 +514,11 @@ i64 DynstrSection<E>::find_string(std::string_view str) {
 template <typename E>
 void DynstrSection<E>::copy_buf(Context<E> &ctx) {
   u8 *base = ctx.buf + this->shdr.sh_offset;
-  base[0] = '\0';
 
   for (std::pair<std::string_view, i64> p : strings)
     write_string(base + p.second, p.first);
 
-  i64 off = dynsym_offset;
+  i64 off = ctx.dynsym->dynstr_offset;
   for (Symbol<E> *sym : ctx.dynsym->symbols)
     if (sym)
       off += write_string(base + off, sym->name());
@@ -667,13 +618,23 @@ void SymtabSection<E>::copy_buf(Context<E> &ctx) {
 // is set in the dynamic section.
 //
 // This function returns true if DT_AARCH64_VARIANT_PCS needs to be set.
-static bool contains_variant_pcs(Context<ARM64> &ctx) {
-  for (Symbol<ARM64> *sym : ctx.plt->symbols)
+template <is_arm64 E>
+static bool contains_variant_pcs(Context<E> &ctx) {
+  for (Symbol<E> *sym : ctx.plt->symbols)
     if (sym->esym().arm64_variant_pcs)
       return true;
   return false;
 }
 
+// RISC-V has the same feature but with a different name.
+template <is_riscv E>
+static bool contains_variant_cc(Context<E> &ctx) {
+  for (Symbol<E> *sym : ctx.plt->symbols)
+    if (sym->esym().riscv_variant_cc)
+      return true;
+  return false;
+}
+
 template <typename E>
 static std::vector<Word<E>> create_dynamic_section(Context<E> &ctx) {
   std::vector<Word<E>> vec;
@@ -825,7 +786,11 @@ static std::vector<Word<E>> create_dynamic_section(Context<E> &ctx) {
 
   if constexpr (is_arm64<E>)
     if (contains_variant_pcs(ctx))
-      define(DT_AARCH64_VARIANT_PCS, 1);
+      define(DT_AARCH64_VARIANT_PCS, 0);
+
+  if constexpr (is_riscv<E>)
+    if (contains_variant_cc(ctx))
+      define(DT_RISCV_VARIANT_CC, 0);
 
   if constexpr (is_ppc32<E>)
     define(DT_PPC_GOT, ctx.gotplt->shdr.sh_addr);
@@ -834,7 +799,7 @@ static std::vector<Word<E>> create_dynamic_section(Context<E> &ctx) {
     // PPC64_GLINK is defined by the psABI to refer 32 bytes before
     // the first PLT entry. I don't know why it's 32 bytes off, but
     // it's what it is.
-    define(DT_PPC64_GLINK, ctx.plt->shdr.sh_addr + E::plt_hdr_size - 32);
+    define(DT_PPC64_GLINK, ctx.plt->shdr.sh_addr + to_plt_offset<E>(0) - 32);
   }
 
   // GDB needs a DT_DEBUG entry in an executable to store a word-size
@@ -846,7 +811,6 @@ static std::vector<Word<E>> create_dynamic_section(Context<E> &ctx) {
 
   for (i64 i = 0; i < ctx.arg.spare_dynamic_tags; i++)
     define(DT_NULL, 0);
-
   return vec;
 }
 
@@ -886,16 +850,10 @@ template <typename E>
 void OutputSection<E>::compute_section_size(Context<E> &ctx) {
   ElfShdr<E> &shdr = this->shdr;
 
-  // On most RISC systems, we need to create so-called "range extension
-  // thunks" to extend branch instructions reach, as their jump
-  // instructions' reach is limited. create_range_extension_thunks()
-  // computes the size of the section while inserting thunks.
-  if constexpr (needs_thunk<E>) {
-    if ((shdr.sh_flags & SHF_EXECINSTR) && !ctx.arg.relocatable) {
-      create_range_extension_thunks(ctx);
-      return;
-    }
-  }
+  // Text sections must to be handled by create_range_extension_thunks()
+  // if they may need range extension thunks.
+  assert(!needs_thunk<E> || !(shdr.sh_flags & SHF_EXECINSTR) ||
+         ctx.arg.relocatable);
 
   // Since one output section may contain millions of input sections,
   // we first split input sections into groups and assign offsets to
@@ -946,52 +904,20 @@ void OutputSection<E>::compute_section_size(Context<E> &ctx) {
 
 template <typename E>
 void OutputSection<E>::copy_buf(Context<E> &ctx) {
-  if (this->shdr.sh_type != SHT_NOBITS) {
-    ElfRel<E> *rel = nullptr;
-    if (ctx.reldyn)
-      rel = (ElfRel<E> *)(ctx.buf + ctx.reldyn->shdr.sh_offset +
-                          this->reldyn_offset);
+  if (this->shdr.sh_type == SHT_NOBITS)
+    return;
 
-    write_to(ctx, ctx.buf + this->shdr.sh_offset, rel);
-  }
-}
+  // Copy section contents
+  u8 *buf = ctx.buf + this->shdr.sh_offset;
+  write_to(ctx, buf);
 
-template <typename E>
-void OutputSection<E>::write_to(Context<E> &ctx, u8 *buf, ElfRel<E> *rel) {
-  // Copy section contents to an output file.
-  tbb::parallel_for((i64)0, (i64)members.size(), [&](i64 i) {
-    InputSection<E> &isec = *members[i];
-    isec.write_to(ctx, buf + isec.offset);
+  // Emit dynamic relocations
+  if (!ctx.reldyn)
+    return;
 
-    // Clear trailing padding. We write trap or nop instructions for
-    // an executable segment so that a disassembler wouldn't try to
-    // disassemble garbage as instructions.
-    u64 this_end = isec.offset + isec.sh_size;
-    u64 next_start;
-    if (i + 1 < members.size())
-      next_start = members[i + 1]->offset;
-    else
-      next_start = this->shdr.sh_size;
+  ElfRel<E> *rel = (ElfRel<E> *)(ctx.buf + ctx.reldyn->shdr.sh_offset +
+                                 this->reldyn_offset);
 
-    u8 *loc = buf + this_end;
-    i64 size = next_start - this_end;
-
-    if (this->shdr.sh_flags & SHF_EXECINSTR) {
-      for (i64 i = 0; i + sizeof(E::filler) <= size; i += sizeof(E::filler))
-        memcpy(loc + i, E::filler, sizeof(E::filler));
-    } else {
-      memset(loc, 0, size);
-    }
-  });
-
-  // Emit range extension thunks.
-  if constexpr (needs_thunk<E>) {
-    tbb::parallel_for_each(thunks, [&](std::unique_ptr<Thunk<E>> &thunk) {
-      thunk->copy_buf(ctx);
-    });
-  }
-
-  // Emit dynamic relocations.
   for (AbsRel<E> &r : abs_rels) {
     Word<E> *loc = (Word<E> *)(buf + r.isec->offset + r.offset);
     u64 addr = this->shdr.sh_addr + r.isec->offset + r.offset;
@@ -1026,6 +952,42 @@ void OutputSection<E>::write_to(Context<E> &ctx, u8 *buf, ElfRel<E> *rel) {
   }
 }
 
+template <typename E>
+void OutputSection<E>::write_to(Context<E> &ctx, u8 *buf) {
+  // Copy section contents to an output file.
+  tbb::parallel_for((i64)0, (i64)members.size(), [&](i64 i) {
+    InputSection<E> &isec = *members[i];
+    isec.write_to(ctx, buf + isec.offset);
+
+    // Clear trailing padding. We write trap or nop instructions for
+    // an executable segment so that a disassembler wouldn't try to
+    // disassemble garbage as instructions.
+    u64 this_end = isec.offset + isec.sh_size;
+    u64 next_start;
+    if (i + 1 < members.size())
+      next_start = members[i + 1]->offset;
+    else
+      next_start = this->shdr.sh_size;
+
+    u8 *loc = buf + this_end;
+    i64 size = next_start - this_end;
+
+    if (this->shdr.sh_flags & SHF_EXECINSTR) {
+      for (i64 i = 0; i + sizeof(E::filler) <= size; i += sizeof(E::filler))
+        memcpy(loc + i, E::filler, sizeof(E::filler));
+    } else {
+      memset(loc, 0, size);
+    }
+  });
+
+  // Emit range extension thunks.
+  if constexpr (needs_thunk<E>) {
+    tbb::parallel_for_each(thunks, [&](std::unique_ptr<Thunk<E>> &thunk) {
+      thunk->copy_buf(ctx);
+    });
+  }
+}
+
 // .relr.dyn contains base relocations encoded in a space-efficient form.
 // The contents of the section is essentially just a list of addresses
 // that have to be fixed up at runtime.
@@ -1181,7 +1143,7 @@ void OutputSection<E>::compute_symtab_size(Context<E> &ctx) {
         this->num_local_symtab += thunk->symbols.size();
 
       for (Symbol<E> *sym : thunk->symbols)
-        this->strtab_size += sym->name().size() + sizeof("$thunk");
+        this->strtab_size += sym->name().size() + thunk->name.size() + 2;
     }
   }
 }
@@ -1215,7 +1177,8 @@ void OutputSection<E>::populate_symtab(Context<E> &ctx) {
         write_esym(addr, strtab - strtab_base);
 
         strtab += write_string(strtab, sym.name()) - 1;
-        strtab += write_string(strtab, "$thunk");
+        *strtab++ = '$';
+        strtab += write_string(strtab, thunk->name);
 
         // Emit "$t", "$a" and "$d" if ARM32.
         if constexpr (is_arm32<E>) {
@@ -1323,12 +1286,6 @@ static std::vector<GotEntry<E>> get_got_entries(Context<E> &ctx) {
   for (Symbol<E> *sym : ctx.got->got_syms) {
     i64 idx = sym->get_got_idx(ctx);
 
-    // If a symbol is imported, let the dynamic linker to resolve it.
-    if (sym->is_imported) {
-      add({idx, 0, E::R_GLOB_DAT, sym});
-      continue;
-    }
-
     // IFUNC always needs to be fixed up by the dynamic linker.
     if constexpr (supports_ifunc<E>) {
       if (sym->is_ifunc()) {
@@ -1342,6 +1299,12 @@ static std::vector<GotEntry<E>> get_got_entries(Context<E> &ctx) {
       }
     }
 
+    // If a symbol is imported, let the dynamic linker to resolve it.
+    if (sym->is_imported) {
+      add({idx, 0, E::R_GLOB_DAT, sym});
+      continue;
+    }
+
     // If we know an address at link-time, fill that GOT entry now.
     // It may need a base relocation, though.
     if (ctx.arg.pic && sym->is_relative())
@@ -1746,6 +1709,20 @@ void RelPltSection<E>::copy_buf(Context<E> &ctx) {
   }
 }
 
+// RISC-V and LoongArch have code-shrinking linker relaxation. If we
+// have removed instructions from a function, we need to update its
+// size as well.
+template <typename E>
+static u64 get_symbol_size(Symbol<E> &sym) {
+  const ElfSym<E> &esym = sym.esym();
+  if constexpr (is_riscv<E> || is_loongarch<E>)
+    if (esym.st_size > 0)
+      if (InputSection<E> *isec = sym.get_input_section())
+        return esym.st_size + esym.st_value - sym.value -
+               get_r_delta(*isec, esym.st_value + esym.st_size);
+  return esym.st_size;
+}
+
 template <typename E>
 ElfSym<E> to_output_esym(Context<E> &ctx, Symbol<E> &sym, u32 st_name,
                          U32<E> *shn_xindex) {
@@ -1754,7 +1731,7 @@ ElfSym<E> to_output_esym(Context<E> &ctx, Symbol<E> &sym, u32 st_name,
 
   esym.st_name = st_name;
   esym.st_type = sym.get_type();
-  esym.st_size = sym.esym().st_size;
+  esym.st_size = get_symbol_size(sym);
 
   if (sym.is_local(ctx))
     esym.st_bind = STB_LOCAL;
@@ -1768,8 +1745,11 @@ ElfSym<E> to_output_esym(Context<E> &ctx, Symbol<E> &sym, u32 st_name,
   if constexpr (is_arm64<E>)
     esym.arm64_variant_pcs = sym.esym().arm64_variant_pcs;
 
+  if constexpr (is_riscv<E>)
+    esym.riscv_variant_cc = sym.esym().riscv_variant_cc;
+
   if constexpr (is_ppc64v2<E>)
-    esym.ppc_local_entry = sym.esym().ppc_local_entry;
+    esym.ppc64_local_entry = sym.esym().ppc64_local_entry;
 
   auto get_st_shndx = [&](Symbol<E> &sym) -> u32 {
     if (SectionFragment<E> *frag = sym.get_frag())
@@ -1881,18 +1861,15 @@ void DynsymSection<E>::update_shdr(Context<E> &ctx) {
 
 template <typename E>
 void DynsymSection<E>::copy_buf(Context<E> &ctx) {
-  u8 *base = ctx.buf + this->shdr.sh_offset;
-  memset(base, 0, sizeof(ElfSym<E>));
-  i64 name_offset = ctx.dynstr->dynsym_offset;
+  ElfSym<E> *buf = (ElfSym<E> *)(ctx.buf + this->shdr.sh_offset);
+  i64 offset = dynstr_offset;
+
+  memset(buf, 0, sizeof(ElfSym<E>));
 
   for (i64 i = 1; i < symbols.size(); i++) {
     Symbol<E> &sym = *symbols[i];
-    ElfSym<E> &esym =
-      *(ElfSym<E> *)(base + sym.get_dynsym_idx(ctx) * sizeof(ElfSym<E>));
-
-    esym = to_output_esym(ctx, sym, name_offset, nullptr);
-    name_offset += sym.name().size() + 1;
-    assert(esym.st_bind != STB_LOCAL || i < this->shdr.sh_info);
+    buf[sym.get_dynsym_idx(ctx)] = to_output_esym(ctx, sym, offset, nullptr);
+    offset += sym.name().size() + 1;
   }
 }
 
@@ -1901,9 +1878,9 @@ void HashSection<E>::update_shdr(Context<E> &ctx) {
   if (ctx.dynsym->symbols.empty())
     return;
 
-  i64 header_size = 8;
+  i64 header_size = sizeof(Entry) * 2;
   i64 num_slots = ctx.dynsym->symbols.size();
-  this->shdr.sh_size = header_size + num_slots * 8;
+  this->shdr.sh_size = header_size + num_slots * sizeof(Entry) * 2;
   this->shdr.sh_link = ctx.dynsym->shndx;
 }
 
@@ -1913,21 +1890,18 @@ void HashSection<E>::copy_buf(Context<E> &ctx) {
   memset(base, 0, this->shdr.sh_size);
 
   std::span<Symbol<E> *> syms = ctx.dynsym->symbols;
-  U32<E> *hdr = (U32<E> *)base;
-  U32<E> *buckets = (U32<E> *)(base + 8);
-  U32<E> *chains = buckets + syms.size();
+  Entry *hdr = (Entry *)base;
+  Entry *buckets = hdr + 2;
+  Entry *chains = buckets + syms.size();
 
-  hdr[0] = hdr[1] = syms.size();
+  hdr[0] = syms.size();
+  hdr[1] = syms.size();
 
-  std::vector<u32> hashes(syms.size());
-  tbb::parallel_for((i64)1, (i64)syms.size(), [&](i64 i) {
-    hashes[i] = elf_hash(syms[i]->name()) % syms.size();
-  });
-
-  for (i64 i = 1; i < syms.size(); i++) {
-    i64 h = hashes[i];
-    chains[syms[i]->get_dynsym_idx(ctx)] = buckets[h];
-    buckets[h] = syms[i]->get_dynsym_idx(ctx);
+  for (Symbol<E> *sym : syms.subspan(1)) {
+    i64 i = sym->get_dynsym_idx(ctx);
+    i64 h = elf_hash(sym->name()) % syms.size();
+    chains[i] = buckets[h];
+    buckets[h] = i;
   }
 }
 
@@ -1954,18 +1928,20 @@ void GnuHashSection<E>::copy_buf(Context<E> &ctx) {
 
   i64 first_exported = ctx.dynsym->symbols.size() - num_exported;
 
-  std::span<Symbol<E> *> syms = ctx.dynsym->symbols;
-  syms = syms.subspan(first_exported);
-
-  std::vector<u32> indices(num_exported);
-
   *(U32<E> *)base = num_buckets;
   *(U32<E> *)(base + 4) = first_exported;
   *(U32<E> *)(base + 8) = num_bloom;
   *(U32<E> *)(base + 12) = BLOOM_SHIFT;
 
+  std::span<Symbol<E> *> syms = ctx.dynsym->symbols;
+  syms = syms.subspan(first_exported);
+
+  if (syms.empty())
+    return;
+
   // Write a bloom filter
   Word<E> *bloom = (Word<E> *)(base + HEADER_SIZE);
+  std::vector<u32> indices(num_exported);
 
   for (i64 i = 0; i < syms.size(); i++) {
     constexpr i64 word_bits = sizeof(Word<E>) * 8;
@@ -1981,9 +1957,8 @@ void GnuHashSection<E>::copy_buf(Context<E> &ctx) {
   // Write hash bucket indices
   U32<E> *buckets = (U32<E> *)(bloom + num_bloom);
 
-  for (i64 i = 0; i < syms.size(); i++)
-    if (!buckets[indices[i]])
-      buckets[indices[i]] = i + first_exported;
+  for (i64 i = syms.size() - 1; i >= 0; i--)
+    buckets[indices[i]] = first_exported + i;
 
   // Write a hash table
   U32<E> *table = buckets + num_buckets;
@@ -2008,14 +1983,22 @@ get_merged_output_name(Context<E> &ctx, std::string_view name, u64 flags,
   if (ctx.arg.unique && ctx.arg.unique->match(name))
     return name;
 
-  // GCC seems to create sections named ".rodata.strN.<mangled-symbol-name>.M".
-  // We want to eliminate the symbol name part from the section name.
-  if ((flags & SHF_STRINGS) && name.starts_with(".rodata.")) {
-    std::string name2 = ".rodata.str"s + std::to_string(entsize) +
-                        "." + std::to_string(addralign);
-    if (name == name2)
-      return name;
-    return save_string(ctx, name2);
+  // GCC seems to create sections named ".rodata.strN.<mangled-symbol-name>.M"
+  // or ".rodata.cst.<mangled-symbol-name.cstN". We want to eliminate the
+  // symbol name part from the section name.
+  if (name.starts_with(".rodata.")) {
+    if (flags & SHF_STRINGS) {
+      std::string name2 = ".rodata.str"s + std::to_string(entsize) +
+                          "." + std::to_string(addralign);
+      if (name == name2)
+        return name;
+      return save_string(ctx, name2);
+    } else {
+      std::string name2 = ".rodata.cst"s + std::to_string(entsize);
+      if (name == name2)
+        return name;
+      return save_string(ctx, name2);
+    }
   }
 
   return name;
@@ -2191,17 +2174,18 @@ void MergedSection<E>::compute_section_size(Context<E> &ctx) {
 
 template <typename E>
 void MergedSection<E>::copy_buf(Context<E> &ctx) {
-  write_to(ctx, ctx.buf + this->shdr.sh_offset, nullptr);
+  write_to(ctx, ctx.buf + this->shdr.sh_offset);
 }
 
 template <typename E>
-void MergedSection<E>::write_to(Context<E> &ctx, u8 *buf, ElfRel<E> *rel) {
+void MergedSection<E>::write_to(Context<E> &ctx, u8 *buf) {
   i64 shard_size = map.nbuckets / map.NUM_SHARDS;
 
   tbb::parallel_for((i64)0, map.NUM_SHARDS, [&](i64 i) {
     // There might be gaps between strings to satisfy alignment requirements.
     // If that's the case, we need to zero-clear them.
-    if (this->shdr.sh_addralign > 1)
+    if (this->shdr.sh_addralign > 1 &&
+        this->shdr.sh_addralign != this->shdr.sh_entsize)
       memset(buf + shard_offsets[i], 0, shard_offsets[i + 1] - shard_offsets[i]);
 
     // Copy strings
@@ -2230,7 +2214,7 @@ void EhFrameSection<E>::construct(Context<E> &ctx) {
 
   // Remove dead FDEs and assign them offsets within their corresponding
   // CIE group.
-  tbb::parallel_for_each(ctx.objs, [&](ObjectFile<E> *file) {
+  tbb::parallel_for_each(ctx.objs, [](ObjectFile<E> *file) {
     std::erase_if(file->fdes, [](FdeRecord<E> &fde) { return !fde.is_alive; });
 
     i64 offset = 0;
@@ -2291,7 +2275,7 @@ void EhFrameSection<E>::copy_buf(Context<E> &ctx) {
   HdrEntry *eh_hdr = nullptr;
   if (ctx.eh_frame_hdr)
     eh_hdr = (HdrEntry *)(ctx.buf + ctx.eh_frame_hdr->shdr.sh_offset +
-                   EhFrameHdrSection<E>::HEADER_SIZE);
+                          EhFrameHdrSection<E>::HEADER_SIZE);
 
   tbb::parallel_for_each(ctx.objs, [&](ObjectFile<E> *file) {
     // Copy CIEs.
@@ -2376,7 +2360,7 @@ template <typename E>
 void EhFrameHdrSection<E>::copy_buf(Context<E> &ctx) {
   u8 *base = ctx.buf + this->shdr.sh_offset;
 
-  // Write a header. The actual table is written by EhFrameHdr<E>::copy_buf.
+  // Write a header. The actual table is written by EhFrameSection::copy_buf.
   base[0] = 1;
   base[1] = DW_EH_PE_pcrel | DW_EH_PE_sdata4;
   base[2] = DW_EH_PE_udata4;
@@ -2745,13 +2729,6 @@ void NotePackageSection<E>::copy_buf(Context<E> &ctx) {
 // Merges input files' .note.gnu.property values.
 template <typename E>
 void NotePropertySection<E>::update_shdr(Context<E> &ctx) {
-  // The rules we support are only specified for x86 psABI
-  if (!is_x86<E>)
-    return;
-
-  // Reset to the initial state so that this function is idempotent
-  properties.clear();
-
   // Obtain the list of keys
   std::vector<ObjectFile<E> *> files = ctx.objs;
   std::erase(files, ctx.internal_obj);
@@ -2769,6 +2746,8 @@ void NotePropertySection<E>::update_shdr(Context<E> &ctx) {
   };
 
   // Merge values for each key
+  std::map<u32, u32> map;
+
   for (u32 key : keys) {
     auto has_key = [&](ObjectFile<E> *file) {
       return file->gnu_properties.contains(key);
@@ -2778,41 +2757,41 @@ void NotePropertySection<E>::update_shdr(Context<E> &ctx) {
         key <= GNU_PROPERTY_X86_UINT32_AND_HI) {
       // An AND feature is set if all input objects have the property and
       // the feature.
-      if (std::all_of(files.begin(), files.end(), has_key)) {
-        properties[key] = 0xffff'ffff;
-        for (ObjectFile<E> *file : files)
-          properties[key] &= get_value(file, key);
-      }
+      map[key] = 0xffff'ffff;
+      for (ObjectFile<E> *file : files)
+        map[key] &= get_value(file, key);
     } else if (GNU_PROPERTY_X86_UINT32_OR_LO <= key &&
                key <= GNU_PROPERTY_X86_UINT32_OR_HI) {
       // An OR feature is set if some input object has the feature.
       for (ObjectFile<E> *file : files)
-        properties[key] |= get_value(file, key);
+        map[key] |= get_value(file, key);
     } else if (GNU_PROPERTY_X86_UINT32_OR_AND_LO <= key &&
                key <= GNU_PROPERTY_X86_UINT32_OR_AND_HI) {
       // An OR-AND feature is set if all input object files have the property
       // and some of them has the feature.
       if (std::all_of(files.begin(), files.end(), has_key))
         for (ObjectFile<E> *file : files)
-          properties[key] |= get_value(file, key);
+          map[key] |= get_value(file, key);
     }
   }
 
   if (ctx.arg.z_ibt)
-    properties[GNU_PROPERTY_X86_FEATURE_1_AND] |= GNU_PROPERTY_X86_FEATURE_1_IBT;
+    map[GNU_PROPERTY_X86_FEATURE_1_AND] |= GNU_PROPERTY_X86_FEATURE_1_IBT;
   if (ctx.arg.z_shstk)
-    properties[GNU_PROPERTY_X86_FEATURE_1_AND] |= GNU_PROPERTY_X86_FEATURE_1_SHSTK;
+    map[GNU_PROPERTY_X86_FEATURE_1_AND] |= GNU_PROPERTY_X86_FEATURE_1_SHSTK;
+  map[GNU_PROPERTY_X86_ISA_1_NEEDED] |= ctx.arg.z_x86_64_isa_level;
 
-  properties[GNU_PROPERTY_X86_ISA_1_NEEDED] |= ctx.arg.z_x86_64_isa_level;
+  // Serialize the map
+  contents.clear();
 
-  std::erase_if(properties, [](std::pair<u32, u32> kv) {
-    return kv.second == 0;
-  });
+  for (std::pair<u32, u32> kv : map)
+    if (kv.second)
+      contents.push_back({kv.first, 4, kv.second});
 
-  if (properties.empty())
+  if (contents.empty())
     this->shdr.sh_size = 0;
   else
-    this->shdr.sh_size = 16 + ENTRY_SIZE * properties.size();
+    this->shdr.sh_size = 16 + contents.size() * sizeof(contents[0]);
 }
 
 template <typename E>
@@ -2820,18 +2799,11 @@ void NotePropertySection<E>::copy_buf(Context<E> &ctx) {
   U32<E> *buf = (U32<E> *)(ctx.buf + this->shdr.sh_offset);
   memset(buf, 0, this->shdr.sh_size);
 
-  buf[0] = 4;                              // Name size
-  buf[1] = ENTRY_SIZE * properties.size(); // Content size
-  buf[2] = NT_GNU_PROPERTY_TYPE_0;         // Type
-  memcpy(buf + 3, "GNU", 4);               // Name
-
-  i64 idx = 4;
-  for (std::pair<u32, u32> kv : properties) {
-    buf[idx] = kv.first;                   // Feature type
-    buf[idx + 1] = 4;                      // Feature size
-    buf[idx + 2] = kv.second;              // Feature flags
-    idx += ENTRY_SIZE / sizeof(U32<E>);
-  }
+  buf[0] = 4;                       // Name size
+  buf[1] = this->shdr.sh_size - 16; // Content size
+  buf[2] = NT_GNU_PROPERTY_TYPE_0;  // Type
+  memcpy(buf + 3, "GNU", 4);        // Name
+  write_vector(buf + 4, contents);  // Content
 }
 
 template <typename E>
@@ -2843,7 +2815,7 @@ CompressedSection<E>::CompressedSection(Context<E> &ctx, Chunk<E> &chunk) {
   this->uncompressed_data.resize(chunk.shdr.sh_size);
   u8 *buf = this->uncompressed_data.data();
 
-  chunk.write_to(ctx, buf, nullptr);
+  chunk.write_to(ctx, buf);
 
   switch (ctx.arg.compress_debug_sections) {
   case COMPRESS_ZLIB:
