@@ -51,15 +51,24 @@ $CC -B. -o $t/exe3 $t/c.o $t/d.o -Wl,-no-relax \
   -Wl,--section-start=.low=0x10000000,--section-start=.high=0x20000000
 $QEMU $t/exe3 | grep -q '42 5'
 
+$GCC -B. -shared -o $t/c.so $t/c.o -Wl,-z,nodlopen
+$CC -B. -o $t/exe4 $t/c.so $t/d.o
+$QEMU $t/exe4 | grep -q '42 5'
+
+
 $GCC -fPIC -mtls-dialect=gnu2 -c -o $t/e.o $t/a.c -mthumb
 $GCC -fPIC -mtls-dialect=gnu2 -c -o $t/f.o $t/b.c -mthumb
 
-$CC -B. -o $t/exe4 $t/e.o $t/f.o
-$QEMU $t/exe4 | grep -q '42 5'
-
-$CC -B. -o $t/exe5 $t/e.o $t/f.o -Wl,-no-relax
+$CC -B. -o $t/exe5 $t/e.o $t/f.o
 $QEMU $t/exe5 | grep -q '42 5'
 
-$CC -B. -o $t/exe6 $t/e.o $t/f.o -Wl,-no-relax \
-  -Wl,--section-start=.low=0x10000000,--section-start=.high=0x20000000
+$CC -B. -o $t/exe6 $t/e.o $t/f.o -Wl,-no-relax
 $QEMU $t/exe6 | grep -q '42 5'
+
+$CC -B. -o $t/exe7 $t/e.o $t/f.o -Wl,-no-relax \
+  -Wl,--section-start=.low=0x10000000,--section-start=.high=0x20000000
+$QEMU $t/exe7 | grep -q '42 5'
+
+$GCC -B. -shared -o $t/e.so $t/e.o -Wl,-z,nodlopen
+$CC -B. -o $t/exe8 $t/e.so $t/f.o
+$QEMU $t/exe8 | grep -q '42 5'
